#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/uio.h>
#include <sysexits.h>

#include "misc.h"


/*          *
 ** MEMORY **
 *          */


void *xmalloc(size_t size)
{
 void *p;

 p = malloc(size);
 if( p == NULL )
    m_err(EX_SOFTWARE, "can not allocate memory");

 return p;
}


void *xcalloc(size_t num, size_t size)
{
 void *p;

 p = calloc(num, size);
 if( p == NULL )
    m_err(EX_SOFTWARE, "can not allocate memory");

 return p;
}


void *xrealloc(void *ptr, size_t sz)
{
 void *p;

 p = realloc(ptr, sz);
 if( p == NULL )
    m_err(EX_SOFTWARE, "can not realloc()");

 return p;
}
 

char *strdup_line(char *s)
{
 char *p, *n;

 if( s == NULL || *s == '\0' ) 
    return NULL;

 p = s;
 while( *p != '\n' && *p != '\0' )
       p++;
 n = (char *)xmalloc(p-s+1);
 strncpy(n, s, (p-s));
 n[(p-s)+1] = '\0';

 return n;
}


char *wstrndup(char *s, size_t len)
{
 char *n;

 if( s == NULL || *s == '\0' ) 
    return NULL;

 if( strlen(s) < len )
    n = strdup(s);
 else {
    n = (char *)xmalloc(len+1);
    strncpy(n, s, len);
    n[len] = '\0';
    }

 return n;
}



/*               *
 ** NETWORK I/O **
 *               */


ssize_t freadn(void *vptr, size_t n, size_t ob, FILE *st)
{
 size_t	nleft, nread;
 char	*ptr;

 ptr = vptr;
 nleft = ob;
 while( nleft > 0 ) {
       nread = fread(ptr, n, nleft, st);
       if( nread == 0 ) {
          if( ferror(st) )
             return -1;
          else
             break;                             /* EOF */
          }
       nleft -= nread;
       ptr   += nread;
       }

 return (ob-nleft);		/* return >= 0 */
}


ssize_t fwriten(void *vptr, size_t n, size_t ob, FILE *st)
{
 size_t	nleft, nwrite;
 char	*ptr;

 ptr = vptr;
 nleft = ob;
 while( nleft > 0 ) {
       nwrite = fwrite(ptr, n, nleft, st);
       if( nwrite == 0 ) {
          if( ferror(st) )
             return -1;
          else
             break;                             /* EOF */
          }
       nleft -= nwrite;
       ptr   += nwrite;
       }

 return (ob-nleft);		/* return >= 0 */
}


/*                  *
 ** ERROR HANDLING **
 *                  */


void m_warn(char *s, ...)
{
 va_list ap;

 va_start(ap, s);
 fprintf(stderr, "warning: ");
 vfprintf(stderr, s, ap);
 fprintf(stderr, "\n");
 va_end(ap);
}


void m_err(int code, char *s, ...)
{
 va_list ap;

 va_start(ap, s);
 fprintf(stderr, "fatal: ");
 vfprintf(stderr, s, ap);
 fprintf(stderr, "\n");
 va_end(ap);

 exit(code);
}
