/* M.Meadele mm@bzero.net Tue May  7 17:40:06 GMT 2002 
 *
 *---
 * Tue May  7 17:27:09 GMT 2002
 * - first version
 * - removed process_strurl()
 * - removed poison_action()
 * - removed process_url()
 *---
 *
 *---
 * Wed May  8 14:57:59 GMT 2002
 * - modified print_list() and print_list_query()
 *
 *---
 * Thu May  9 14:27:01 GMT 2002
 * - corrected bug in list_add_query()
 * - added poison_analyse_content()
 * - added config signature file handling (section SIGNATURE MAKER)
 *
 * ---
 * Sat May 11 19:23:22 GMT 2002
 * - moved main() from htmlparser.l to poison.c
 *  
 * ---
 * Fri May 24 03:43:57 GMT 2002
 * - added cookies.h
 * - added cookie function in main()
 *
 * ---
 * Sat May 25 22:44:51 GMT 2002
 * - added poison_target()
 * - added stupid redirection
 * - added poison_print_report()
 * - removed global ul_list *ul
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sysexits.h>
#include <signal.h>

#include <errno.h>

#include <regex.h>

#include "misc.h"
#include "url.h"
#include "cookies.h"
#include "http.h"
#include "htmlparser.h"

#include "poison.h"


static void poison_usage(char *);

static struct res *is_res_member(char *);
static struct query *is_query_member(struct res *, char *);
static void list_free_query(struct q_list *);
static struct res *list_add_res(char *);
static struct query *list_add_query(struct res *, char *);
static void list_print_query(struct q_list *);

static int poison_analyse(struct http_msg *, struct w_sign *);
static int poison_analyse_content(struct http_msg *, struct w_sign *);
static int poison_url(struct host_info *, struct s_url *, struct w_sign *);
static int poison_url_arg(struct host_info *, struct s_url *, struct s_param *, struct w_sign *);
static void poison_print_report(struct url_list *);

static struct w_sign *sig_process_config(char *);
static void sig_process_line(char *, struct w_sign *);
static void sig_free(struct w_sign *);
#ifdef WDEBUG
static void sig_print(struct w_sign *);
#endif
static void sig_add(struct w_sign *, char *);
static int poison_target(char *);



static struct res_list *rlist;
static struct host_info *root_info;
static struct s_url     *root_url;

int nline = 1;

/*
 * test
 */
#ifdef WTEST
char *stressos[] =
{
 "hello_wpoison_test"
 };
#else
char *stressos[] = 
{
 "",
 "\'bad_bad_value",
 "bad_bad_value\'",
 "9, 9, 9",
 "\' OR \'",
 "\'",
 "\'\""
 };
#endif 




void w_sigint(int no)
{

 printf("\n\n** interrupt\n");
 list_print();

 exit(0);
}


int poison_stress_test(struct url_list *ul, struct w_sign *k)
{
 int i;
 struct e_url *cu;
 struct sigaction wnew, wold;

 wnew.sa_handler = w_sigint;
 sigemptyset(&wnew.sa_mask);
 sigaction(SIGINT, &wnew, &wold);

 list_init();

 cu = ul->head;
 for(i=0; i<ul->sz; i++, cu=cu->next)  {
     if( url_recalibrate(cu->url, root_url) == -1 ) {
        m_warn("%s", url_strerror(url_errno));
        continue;
        }
     if( cu->url->query->sz == 0 ) {
        fprintf(stdout, "+ %s is a static document (skipping)\n", cu->url->path);
        continue;
        }
     poison_url(root_info, cu->url, k);
     }
 
 poison_print_report(ul);
#ifdef WDEBUG
 list_print();
#endif
 list_free();

 if( k != NULL )
    sig_free(k);

 return 0;
}


/*              *
 ** LIST STUFF **
 *              */


static struct res *is_res_member(char *s)
{
 int i;
 struct res *c;

 c = rlist->head;

 for(i=0; c != NULL; i++, c=c->next) {
     if( strcmp(c->name, s) == 0 )
        return c;
     }     

 return NULL;
}


static struct query *is_query_member(struct res *r, char *s)
{
 int i;
 struct query *q;

 q = r->qlist->head;

 for(i=0; q != NULL; i++, q=q->next) {
     if( strcmp(q->name, s) == 0 )
        return q;
     }     

 return NULL;
}


static void list_free_query(struct q_list *q)
{
 int i;
 struct query *h, *n;

 h = q->head;
 for(i=0; i<q->sz; i++) {
     n = h->next;
     free(h->name);
     free(h); 
     h = n;
     }
 free(q);
}


void list_free()
{
 int i;
 struct res *h, *n;

 h = rlist->head;
 for(i=0; i<rlist->sz; i++) {
     n = h->next;
     free(h->name);
     list_free_query(h->qlist);
     free(h); 
     h = n;
     }
 free(rlist);
}

 
static struct res *list_add_res(char *s)
{
 struct res *t;
 struct res_list *r;

 r = rlist;
 
 if( r->head == NULL ) { 
    t = r->head = xmalloc(sizeof(struct res));
    r->head->name = strdup(s);
    r->head->qlist = xcalloc(1, sizeof(struct q_list));
    r->head->nvul = 0;
    r->head->next = NULL;
    r->tail = r->head;
    }
 else {
    t = r->tail->next = xmalloc(sizeof(struct res));
    r->tail->next->name = strdup(s);
    r->tail->next->qlist = xcalloc(1, sizeof(struct q_list));
    r->tail->next->nvul = 0;
    r->tail->next->next = NULL;
    r->tail = r->tail->next;
    }
 r->sz++;

 return t;
}


static struct query *list_add_query(struct res *res, char *s)
{
 struct query *t;
 struct q_list *r;

 r = res->qlist;
 
 if( r->head == NULL ) { 
    t = r->head = xmalloc(sizeof(struct query));
    r->head->name = strdup(s);
    r->head->status = 0;
    r->head->next = NULL;
    r->tail = r->head;
    }
 else {
    t = r->tail->next = xmalloc(sizeof(struct query));
    r->tail->next->name = strdup(s);
    r->tail->next->status = 0;
    r->tail->next->next = NULL;
    r->tail = r->tail->next;
    }
 r->sz++;

 return t;
}


static void list_print_query(struct q_list *q)
{
 int i, j;
 struct query *c;

 c = q->head;
 for(i=0; i<q->sz; i++, c=c->next) {
     printf("  |___ %s ", c->name);
     for(j=0; j<c->status; j++)
         printf("* ");
     printf("\n");
     }
}


void list_print()
{
 int i;
 struct res *c;

 c = rlist->head;
 for(i=0; i<rlist->sz; i++, c=c->next) {
     printf(" __ %s    ", c->name);
     if( c->nvul > 0 )
        printf("(%d) [Possible SQL-injection detected]", c->nvul);
     printf("\n");
     list_print_query(c->qlist);
     }
}


void list_init()
{
 
 rlist = xcalloc(1, sizeof(struct res_list));
}


/*          *
 ** POISON **
 *          */

/*
 * test
 */
static int poison_analyse(struct http_msg *msg, struct w_sign *w)
{
 int r;

 /*
  * analyse reply code
  * 500 reply code is often send 
  * by vulnerable asp code,
  * but, it should be analysed deeper.
  */
 r = http_get_rcode(msg);

 switch( r ) {
    case 404:
       break;
    case 500:
       return 1;    /* good ? */
    case 302:
    case 200:
       /*
        * analyse headers + content
        * poison_analyse_header() not yet implemented
        */
       if( w != NULL && msg->content_len )
          return (poison_analyse_content(msg, w));
       break;
    default:
       fprintf(stderr, "strange reply (%d)\n", r);
       break;
    }

 return 0;
}


/*
 * search for typical error keywords in content
 */

/*
 * XXX: severe performance problem
 *      stupid loop
 */
static int poison_analyse_content(struct http_msg *msg, struct w_sign *w)
{
 int i, r;
 struct key_sig *c;

 c = w->head;
  
 for(i=0; i<w->sz; i++, c=c->next) {
     r = regexec(&(c->expr), msg->content, 0, NULL, REG_NOTEOL);
     if( r == 0 ) {
#ifdef WDEBUG
        printf("DEBUG: pattern found: %s\n", c->sig);
#endif
        return 1;
        }
     }

 return 0;
}


static int poison_url(struct host_info *hi, struct s_url *url, struct w_sign *k)
{
 int i, d;
 struct res *c;

 d = 0;

 c = is_res_member(url->path);
 if( c == NULL ) {
    printf("\npoison on: %s\n", url->path);
    printf("poisoning arguments:\n    ");
    for(i=0; i<url->query->sz; i++) 
        printf("<%s> ", url->query->pair[i]->name);
    printf("\n");
    c = list_add_res(url->path);
    for(i=0; i<url->query->sz; i++) {
        list_add_query(c, url->query->pair[i]->name);
        c->qlist->tail->status = poison_url_arg(hi, url, url->query->pair[i], k);
        if( c->qlist->tail->status )
           c->nvul++;
        d++;
        }
    }
 else {
    for(i=0; i<url->query->sz; i++) {
        if( is_query_member(c, url->query->pair[i]->name) == NULL ) {
           printf("\npoisoning new argument <%s> for %s\n", url->query->pair[i]->name, url->path);
           list_add_query(c, url->query->pair[i]->name);
           c->qlist->tail->status = poison_url_arg(hi, url, url->query->pair[i], k);
           if( c->qlist->tail->status )
              c->nvul++;
           d++;
           }
        }
    }

 return d;
}
     

static int poison_url_arg(struct host_info *hi, struct s_url *url, struct s_param *s, struct w_sign *k)
{
 char *bk;
 char *sturl;
 int i, r, tr;
 struct http_msg *rep;

 printf("processing: %s\n", s->name);

 bk = s->value;

 tr = r = 0;
 for(i=0; i<sizeof(stressos)/sizeof(char *); i++) {
     s->value = stressos[i];
  /*   http_set_method(url->method); */
     rep = http_action(hi, url);
#ifdef WDEBUG
     printf("reply line: %s\n", rep->st_line);
#endif
     if( rep == NULL ) {
        fprintf(stderr, "(%s)\n", http_strerror(http_errno));
        continue;
        }
     r = poison_analyse(rep, k);
     tr += r;
     sturl = url_tostr(url);
     printf("%s %s\n", (tr ? "++" : "--"), sturl);
     free(sturl);
     http_free_msg(rep);
     }
 printf("\n");
 s->value = bk;

 return tr;
}


static void poison_print_report(struct url_list *ul)
{
 int i, v;
 struct res *c;

 printf("[***] report:\n");
 printf("    %d links tested:\n", ul->sz);
 c = rlist->head;
 for(v=0, i=0; i<rlist->sz; i++, c=c->next) {
     if( c->nvul > 0 ) {
        printf(" __ %s    ", c->name);
        printf("[Possible SQL-injection detected]\n");
        list_print_query(c->qlist);
        v++;
        }
     }
 if( v == 0 ) 
    printf("no problem found\n");
 else 
    printf("%d potential security problems found\n", v);
}


/*                   *
 ** SIGNATURE MAKER **
 *                   */


#define LINEMAXLEN 1024


static void sig_add(struct w_sign *c, char *s)
{
 int r;
 char m_errbuf[512];

 if( c->head == NULL ) {
    c->head = xmalloc(sizeof(struct key_sig));
    c->head->sig = strdup(s);
    r = regcomp(&(c->head->expr), c->head->sig, REG_EXTENDED|REG_ICASE);
    if( r != 0 ) {
       regerror(r, &(c->head->expr), m_errbuf, sizeof(m_errbuf));
       m_err(EX_USAGE, "line %d: %s", nline, m_errbuf);
       }
    c->head->next = NULL;
    c->tail = c->head;
    }
 else {
    c->tail->next = xmalloc(sizeof(struct key_sig));
    c->tail->next->sig = strdup(s);
    r = regcomp(&(c->tail->next->expr), c->tail->next->sig, REG_EXTENDED|REG_ICASE);
    if( r != 0 ) {
       regerror(r, &(c->tail->next->expr), m_errbuf, sizeof(m_errbuf));
       m_err(EX_USAGE, "line %d: %s", nline,  m_errbuf);
       }
    c->tail->next->next = NULL;
    c->tail = c->tail->next;
    }
 c->sz++;
}


#ifdef WDEBUG
static void sig_print(struct w_sign *c)
{
 int i;
 struct key_sig *k;

 k = c->head;
 printf("\n** signatures:\n\n");
 for(i=0; i<c->sz; i++, k=k->next) 
     printf("%d: %s\n", i, k->sig);

 printf("\n** %d records\n", i);
}
#endif


static void sig_free(struct w_sign *c)
{
 int i;
 struct key_sig *k, *p;

 k = c->head;
 for(i=0; i<c->sz; i++) {
     p = k->next;
     free(k->sig);
     regfree(&(k->expr));
     free(k);
     k = p;
     } 
   
 free(c);
}


static void sig_process_line(char *buf, struct w_sign *c)
{
 char *p, *name;

 p = buf;

 SKIPSP(p);
 if( *p == '#' || *p == '\n' )
    return;

 name = p;

 GOTOLF(p);
 while( *(p-1) == ' ' )
       p--;
 *p = '\0';

 sig_add(c, name);
 return;
}


static struct w_sign *sig_process_config(char *path)
{
 struct w_sign *ws;
 FILE *fp;
 char  buff[LINEMAXLEN];

 fp = fopen(path, "r");
 if( fp == NULL ) 
    return NULL;

 ws = xcalloc(1, sizeof(struct w_sign));

 for(;;) {
   if( fgets(buff, LINEMAXLEN-1, fp) == NULL ) {
      if( feof(fp) ) {
         fclose(fp);
         return ws;
         }
      else 
         m_err(EX_SOFTWARE, "%s", strerror(errno));
      }
   sig_process_line(buff, ws);
   nline++;
   }

 fclose(fp);
 return NULL;
}


/*        *
 ** MAIN **
 *        */


static void poison_usage(char *p)
{
 
 fprintf(stderr, "usage:\n");
 fprintf(stderr, "%s [-R [level]] [-C <cookie value>] [-o file]\n", p);
 fprintf(stderr, "          [-f signature file] url\n");
 fprintf(stderr, "   -R [level]        :  Recursive scan, the level value (integer)\n");
 fprintf(stderr, "                        means the maximum level of recursion,\n"); 
 fprintf(stderr, "                        a level of 0, or no level,  means to scan the entire tree.\n");
 fprintf(stderr, "                      (this is not yet implemented).\n");
 fprintf(stderr, "   -C <cookie value> : force value of cookie, usefull for\n"); 
 fprintf(stderr, "                       web-sites which deal with authentification\n");
 fprintf(stderr, "                       <cookie value> is in the form \"name=value\".\n");
 fprintf(stderr, "                       This value is the value of the \"Set-Cookie\" header,\n");
 fprintf(stderr, "                       not the \"Cookie\" one.\n");
 fprintf(stderr, "                       you can add more than one cookie by adding more -C opt.\n");
 fprintf(stderr, "   -o file           : print result to the specified file, in html format.\n");
 fprintf(stderr, "                      (this is not yet implemented).\n");
 fprintf(stderr, "   -f signature file : specifies the path of the signature file,\n");
 fprintf(stderr, "                       default is ./poison.sig .\n");
 fprintf(stderr, "   url               : url to fetch web-document from.\n");

 exit(EX_USAGE);
}


static int poison_target(char *t)
{

 /*
  * build url from argument
  */
 root_url = url_build(t);
 if( root_url == NULL ) {
    m_warn("%s", url_strerror(url_errno));
    return -1;
    }

 /*
  * resolve host..
  */
 root_info = get_hostinfo(root_url);
 if( root_info == NULL ) {
    m_warn("%s", http_strerror(http_errno));
    return -1;
    }

 return 0;
}


int main(int argc, char *argv[])
{
 int r, redir;
 int ch;
 int force_cookie = 0;
 struct http_msg  *msg;
 struct w_sign *k;
 struct url_list *ul;
 struct s_url *redir_url; 
 char *pt;
 char *sig_file = SIGNATURE_FILE;

 redir = 0;
 optind = 1;
 while( (ch = getopt(argc, argv, "hR:C:o:f:")) != -1 ) {
       switch( ch ) {
              case 'h':
                 poison_usage(argv[0]);
              case 'R':
                 fprintf(stderr, "sorry this is not yet implemented\n");
                 exit(EX_USAGE);
              case 'C':
                 if( cookie_tree == NULL ) {
                    cook_tree_init();
                    force_cookie = 1;
                    }
                 cookie_insert(optarg, cookie_tree);
                 break;
              case 'o':
                 fprintf(stderr, "sorry this is not yet implemented\n");
                 exit(EX_USAGE);
              case 'f':
                 if( optarg == NULL )
                    poison_usage(argv[0]);
                 sig_file = optarg;
                 break;
              case '?':
              default:
                poison_usage(argv[0]);
              }
       }
 if( argv[optind] == NULL ) 
    poison_usage(argv[0]);

 /*
  * read signature file
  */
 printf("[*] reading signature file...\n");
 k = sig_process_config(sig_file);
 if( k == NULL ) {
    m_warn("no such file %s ", sig_file);
    m_warn("reply content will not be analysed (this is bad)");
    }
 else {
#ifdef WDEBUG
    sig_print(k);
#endif
    printf("%d entries loaded\n", k->sz);
    }

 url_set_context(USER_CONTEXT);
 if( poison_target(argv[optind]) < 0 )
    m_err(EX_SOFTWARE, "poison_target() failed");

 /*
  * retrieve document
  */
 http_init();
 if( force_cookie == 0 )
    cook_tree_init();

 printf("[*] retrieving %s\n", argv[optind]);
 do {
     msg = http_action(root_info, root_url);
     if( msg == NULL )
        m_err(EX_SOFTWARE, "%s\n", http_strerror(http_errno));
     r = http_get_rcode(msg);
     switch( r ) {
            case 200:
                redir = 0;
                break;
            case 301:
            case 302:
            case 303:
            case 307:
               redir = 1;
               pt = http_get_hdr_value(msg, "location");
               if( pt == NULL )
                  m_err(EX_DATAERR, "no \"Location\" header found on %d redirection reply", r);
               m_warn("redirection, rebuilding url from <%s>", pt);
               redir_url = url_redirect(root_url, pt);
               if( redir_url == NULL )
                  m_err(EX_SOFTWARE, "%s\n", url_strerror(url_errno));
               url_free(root_url);
               root_url = redir_url;
               hostinfo_free(root_info);
               root_info = get_hostinfo(root_url);
               if( root_info == NULL )
                  m_err(EX_SOFTWARE, "%s\n", http_strerror(http_errno));
               http_free_msg(msg);
               break;
            default:
               m_err(EX_SOFTWARE, "unexpected reply code :\n    %s\n", msg->st_line);
            }
     } while( redir );

 if( force_cookie == 0 )
    http_cookie_insert(msg, cookie_tree);

 /*
  * get urls from document
  */
 url_set_context(HREF_CONTEXT);
 printf("[*] parsing content...\n");
 ul = w_html_parse(msg->content, msg->content_len+2);
 url_list_urladd(ul, url_dup(root_url));
#ifdef WDEBUG
 url_list_print(ul);
#endif
 http_free_msg(msg);

 /*
  * do the job
  */
 printf("[**] stressing:\n"); 
 poison_stress_test(ul, k);
 printf("[**] done\n");

 url_free(root_url);
 hostinfo_free(root_info);
 url_list_free(ul);
 cook_tree_destroy(cookie_tree);

 return 0;
}

