/* M.Meadele mm@bzero.net Mon May  6 13:33:35 GMT 2002 
 *
 *---
 * Tue May  7 17:27:09 GMT 2002
 * -first version
 * -added method in struct s_url
 * -modified url_tostr();
 * -modified url_recalibrate();
 * -corriged stupid bug in urlencode ( & 0xf0 )
 * -modified url_get_scheme()
 * -modified url_build();
 *---
 *
 *---
 * Mon May 27 03:20:54 GMT 2002
 * - added url_redirect()
 *
 *---
 * Tue May 28 21:17:18 GMT 2002
 * - added url_dup()
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>

#include "misc.h"
#include "http.h"
#include "url.h"

static int url_get_type(struct s_url *);
static int url_get_fragment(char **, struct s_url *);
static int url_get_scheme(char **, struct s_url *);
static int url_get_auth(char **, struct s_url *);
static int url_get_host(char **, struct s_url *);
static int url_get_port(char **, struct s_url *);
static int url_get_path(char **, struct s_url *);
static int url_get_query(char **, struct s_url *);
static int url_query_build(char **, struct s_url *);

static char *url_decode(char *);
static char *url_encode(char *);


static int url_context = USER_CONTEXT;


char *url_strerror(int num)
{
 static char u_strerror[255];

 switch(num) {
    case 0:
      strcpy(u_strerror, "0- no error");
      break;
    case 1:
      strcpy(u_strerror, "1- no string to parse");
      break;
    case 2:
      strcpy(u_strerror, "2- trailing '%' in url (string may be already decoded)");
      break;
    case 3:
      strcpy(u_strerror, "3- unsupported scheme");
      break;
    case 4:
      strcpy(u_strerror, "4- missing host value");
      break;
    case 5:
      strcpy(u_strerror, "5- bad user/pass scheme");
      break;
    case 6:
      strcpy(u_strerror, "6- bad port value");
      break;
    case 7:
      strcpy(u_strerror, "7- query too large");
      break;
    case 8:
      strcpy(u_strerror, "7- forbidden link redirection");
      break;
    case 255:
    default:
      strcpy(u_strerror, "- undefined error");
      break;
    }

 return u_strerror;
}



/*          *
 ** MEMORY **
 *          */


struct s_url *url_alloc()
{
 struct s_url *url;

 url = xcalloc(1, sizeof(struct s_url));
 url->query = xcalloc(1, sizeof(struct s_query));

 return url;
}


void url_free(struct s_url *url)
{
 int i;
  
 if( url->user )
    free(url->user);
 if( url->pass )
    free(url->pass);
 if( url->host )
    free(url->host);
 if( url->path )
    free(url->path);
 for(i=0; i < url->query->sz; i++) {
     free(url->query->pair[i]->name);
     free(url->query->pair[i]->value);
     free(url->query->pair[i]);
     }
 free(url->query);
 if( url->frag )
    free(url->frag);
 free(url);
}


struct url_list *url_list_alloc()
{
 struct url_list *ul;

 ul = xcalloc(1, sizeof(struct url_list));
 
 return ul;
}


void url_list_free(struct url_list *ul)
{
 int i;
 struct e_url *c, *d;

 c = ul->head;
 for(i=0; i<ul->sz; i++) {
     d = c->next;
     e_url_free(c);
     c = d;
     }
 free(ul);
}


struct e_url *e_url_alloc()
{
 struct e_url *el;

 el = xcalloc(1, sizeof(struct e_url));
 
 return el;
}


void e_url_free(struct e_url *el)
{
 
 url_free(el->url);
 free(el);
}


/*         *
 ** PRINT **
 *         */


void url_print(struct s_url *url)
{
 int i;
 struct s_param *c;
  
 printf("[*] start of record\n");
 printf("context: %d\n", url_get_context());
 printf("proto: %d\n", url->scheme);
 if( url->method == 1 )
    printf("method: POST\n");
 else 
    printf("method: GET\n");
 if( url->user )
    printf("user: %s\n", url->user);
 if( url->pass )
    printf("pass: %s\n", url->pass);
 if( url->host )
    printf("host: %s\n", url->host);
 if( url->port )
    printf("port: %d\n", url->port);
 if( url->path )
    printf("path: %s\n", url->path);
 for(i=0; i<url->query->sz; i++) {
     c = url->query->pair[i];
     printf(" query pair %d: %s = %s\n", i,
            ((c->name == NULL) ? "(null)" : c->name),
            ((c->value == NULL) ? "(null)" : c->value));
     }
 if( url->frag )
    printf("fragment: %s\n", url->frag);
 printf("[*] end of record\n");
}


void url_list_print(struct url_list *ul)
{
 int i;
 struct e_url *c;

 c = ul->head;
 for(i=0; i<ul->sz; i++, c=c->next) {
     url_print(c->url);
     printf("\n");
     }
}


/*                 *
 ** DECODE/ENCODE **
 *                 */


static char *
   url_decode(char *url)
{
 int i, j;
 int len, dlen;
 char *durl;

 len  = strlen(url);
 durl = xmalloc(len+1);
 dlen = 0;

 for(i=0; i<len; i++) {
     j = i;
     switch( url[i] ) {
       case '+':
          durl[dlen] = ' ';
          break;
       case '%':
          if( i+2 >= len ) {
             durl[dlen] = url[i];
             break;
             }
          if( !isxdigit(url[i+1]) ||
              !isxdigit(url[i+2])   ) {
             durl[dlen] = url[i];
             break;
             }
          durl[dlen]  = (url[j+1] >= 'A' ?
                        ((url[j+1] & 0xdf) - 'A')+10 : (url[j+1] - '0'));
          durl[dlen]  = durl[dlen] << 4;
          durl[dlen] += (url[j+2] >= 'A' ?
                        ((url[j+2] & 0xdf) - 'A')+10 : (url[j+2] - '0'));
          i += 2;
          break;
       default:
          durl[dlen] = url[i];
          break;
       }
     dlen++;
     }
 durl[dlen] = '\0';

 return durl;
}


static char *url_encode(char *url)
{
 int i;
 int len, elen;
 char *hexn = "0123456789abcdef";
 char *eurl;

 len  = strlen(url);
 eurl = xmalloc((len*3)+1);
 elen = 0;

 for(i=0; i<len; i++) {
     switch( url[i] ) {
       case ' ':    
          eurl[elen] = '+';
          break; 
       default:
          if( (url[i] >= 'a' && url[i] <= 'z') ||
              (url[i] >= 'A' && url[i] <= 'Z') ||
              (url[i] >= '0' && url[i] <= '9') ||
              url[i] == '-' || url[i] == '_' || url[i] == '.' ||
              url[i] == '!' || url[i] == '~' || url[i] == '*' ||
              url[i] == '\'' || url[i] == '(' || url[i] == ')'  ) {
             eurl[elen] = url[i];
             break;
             }
          else {
            eurl[elen++] = '%';
            eurl[elen++] = hexn[(url[i] & 0xf0) >> 4];
            eurl[elen] = hexn[url[i] & 0x0f];
            break;
            }
       }
     elen++;
     }
 eurl[elen] = '\0';

 return eurl;
}




/*          *
 ** PARSER **
 *          */


static int
   url_get_fragment(char **s, struct s_url *url)
{
 char *p;

 p = strchr(*s, '#');
 if( p == NULL )
    return 0;

 *p++ = '\0';
 url->frag = strdup(p);

 return 1;
}


static int
   url_get_scheme(char **s, struct s_url *url)
{
 char *p, *q;

 p = *s;

 url->scheme = 0;   /* not set */

 if( *p == '/' )
    return url->scheme;

 q = strstr(p, "://");
 if( q == NULL ) {
    if( strncmp(p, "mailto:", 7) == 0 )
       url->scheme = PROTO_UNSUPPORTED;
    if( strncmp(p, "news:", 5) == 0 )
       url->scheme = PROTO_UNSUPPORTED;
    return url->scheme;
    }

 if( strncmp(p, "http://", 7) == 0 ) {
    url->scheme = PROTO_HTTP; 
    *s += 7;
    }
 else {
    if( strncmp(p, "https://", 8) == 0 ) {
       url->scheme = PROTO_HTTPS;
       *s += 8;
       }
    else 
       url->scheme = PROTO_UNSUPPORTED;
    }
 
 return url->scheme;
}   


static int
   url_get_type(struct s_url *url)
{

 switch( url->scheme ) {
   case 0:
      if( url_context == HREF_CONTEXT )
         url->type = RELATIVE_URL;
      else
         url->type = NETWORK_URL;
      break;
   default:
      url->type = ABSOLUTE_URL; 
      break;
   }

 return url->scheme;
}


static int
   url_get_auth(char **s, struct s_url *url)
{
 int i;
 char *u, *p, *e;

 u = *s;
 p = e = NULL;

 for(i=0; (u[i] != '\0' && u[i] != '/') && (!p || !e); i++) {
     switch( u[i] ) {
       case ':':
          p = *s+i;
          break;
       case '@':
          e = *s+i;
          break;
       default:
          break;
       }
     }

 if( (e && !p) || ( (p && e) && (e < p)) )
    return -1;

 if( p && e ) {
    *p++ = '\0';
    *e   = '\0';
    url->user = url_decode(u);
    url->pass = url_decode(p);
    *s = e+1;
    return 1;
    }                
 return 0;
}


static int
   url_get_host(char **s, struct s_url *url)
{
 char *e;

 e = strpbrk(*s, ":/");
 if( e == NULL ) {
    url->host = strdup(*s);
    while( **s != '\0' )
          (*s)++;
    return 1;
    }
 if( e == *s )
    return 0;
 url->host = wstrndup(*s,  (e-(*s)));
 *s = e;

 return 1;
}


static int
   url_get_port(char **s, struct s_url *url)
{
 char *t = NULL;

 if( **s != ':' ) {
    url->port = HTTP_DEFAULT_PORT;
    return 0;
    }
 t = strchr(*s, '/');
 if( t == NULL ) {
    url->port = atoi((*s)+1);
    while( **s != '\0' )
          (*s)++;
    if( url->port <= 0 || url->port > 65535 )
       return -1;
    return 1;
    }
 *t = '\0';
 url->port = atoi((*s)+1);
 if( url->port <= 0 || url->port > 65535 )
    return -1;
 *t = '/';
 *s = t;
 
 return 1;
}


static int
   url_get_path(char **s, struct s_url *url)
{
 char *e;

 if( **s == '\0' ) {
    url->path = strdup("/");
    return 0;
    }

 e = strchr(*s, '?');
 if( e == NULL ) {
    url->path = strdup(*s);
    while( **s != '\0' )
          (*s)++;
    return 1;
    }
 *e = '\0';
 url->path = strdup(*s);
 *e = '?';
 *s = e;

 return 1;
}
 

void url_querypair_add(char *n, char *v, int i, struct s_url *url)
{
 struct s_param *c;
 
 c = xmalloc(sizeof(struct s_param));
 c->name  = url_decode(n);
 c->value = url_decode(v);
 url->query->pair[i] = c;
 url->query->sz++;
}   

 
static int
   url_query_build(char **s, struct s_url *url)
{
 int i;
 char *value, *eor;
 
 for(i=0; **s != '\0' && i<MAX_QUERYPAIR; i++) {
     eor = strchr(*s, '&');
     if( eor == NULL ) {
        eor = *s;
        while( *eor != '\0' )
              eor++;
        value = strchr(*s, '=');
        if( value == NULL )
           value = eor;
        else
           *value++ = '\0';
        url_querypair_add(*s, value, i, url);
        *s = eor;
        return ++i;
        }
     *eor = '\0';
     value = strchr(*s, '=');
     if( value == NULL )
        value = eor;
     else
        *value++ = '\0';
     url_querypair_add(*s, value, i, url);
     *s = ++eor;
     }

 if( i == MAX_QUERYPAIR )
    return -1;
 else
    return i;
} 


static int
   url_get_query(char **p, struct s_url *url)
{

 if( **p != '?' ) 
    return 0;
 else {
    *p = *p+1;
    return (url_query_build(p, url));
    }
}


/*           *
 ** BUILDER **
 *           */


struct s_url *url_build(char *purl)
{ 
 int r;
 char *cps, *scps;
 struct s_url *url;

 if( purl == NULL || *purl == '\0' ) {
    url_errno = UE_NOSTRING;
    return NULL;
    }

 url = url_alloc();
 scps = cps = strdup(purl);
 
 /* be odd webmasters compliant */
 while( *cps == ' ' )
        cps++;

 url->method = M_GET; /* default */
 r = url_get_scheme(&cps, url);
 if( r == PROTO_UNSUPPORTED ) {
    url_free(url);
    free(scps);
    url_errno = UE_UNSUPPORTED;
    return NULL;
    }

 url_get_fragment(&cps, url);
 url_get_type(url);

 /* force scheme */
 if( url->scheme == 0 )
    url->scheme = PROTO_HTTP;
 switch( url->type ) {
   case ABSOLUTE_URL:
   case NETWORK_URL:
       r = url_get_auth(&cps, url);
       if( r == -1 ) {
          url_free(url);
          free(scps);
          url_errno = UE_BADLP;
          return NULL;
          }
       r = url_get_host(&cps, url);
       if( r == 0 ) {
          url_free(url);
          free(scps);
          url_errno = UE_NOHOST;
          return NULL;
          }
       r = url_get_port(&cps, url);
       if( r == -1 ) {
          url_free(url);
          free(scps);
          url_errno = UE_BADPORT;
          return NULL;
          }
   case RELATIVE_URL:
       url_get_path(&cps, url);
       r = url_get_query(&cps, url);
       if( r == -1 ) {
          url_free(url);
          free(scps);
          url_errno = UE_QUERYLEN;
          return NULL;
          }
       break;
   default:
       url_free(url);
       free(scps);
       url_errno = UE_UNDEFINED;
       return NULL;
   }       
 free(scps);

 return url;
}


/*
 * sux0r code
 */
char *url_tostr(struct s_url *url)
{
 int plen;
 char *qry, *schm;
 char *u_str, *p, *t;
 
 schm = NULL;
 plen = 0;
 switch( url->type ) {
   case RECALIBRATED_URL:
   case ABSOLUTE_URL:
       switch( url->scheme ) {
         case PROTO_HTTP:
            plen += 7;
            schm = "http://";
            break;
         case PROTO_HTTPS:
            plen += 8;
            schm = "https://";
            break;
         default:
             url_errno = UE_UNDEFINED;
             return NULL;
         }
   case NETWORK_URL:
       if( url->user )
          plen += (strlen(url->user) *3) +1;
       if( url->pass )
          plen += (strlen(url->pass) *3) +1;
       plen += strlen(url->host);
       if( url->port != HTTP_DEFAULT_PORT )
          plen += 5 +1;
       plen++;
   case RELATIVE_URL:
       plen += strlen(url->path);
       qry = url_get_query_string(url);
       if( qry != NULL )
          plen += (strlen(qry) *3) +1;
       plen++;
       break;
   default:
       url_errno = UE_UNDEFINED;
       return NULL;
   }

 u_str = xmalloc(plen);
 p = u_str;      

 switch( url->type ) {
   case RECALIBRATED_URL:
   case ABSOLUTE_URL:
      if( url->scheme != 0 ) 
         p += sprintf(p, "%s", schm);
   case NETWORK_URL:
      if( url->user != NULL ) {
         t = url_encode(url->user);
         p += sprintf(p, "%s:", t);
         free(t);
         if( url->pass != NULL ) {
            t = url_encode(url->pass);
            p += sprintf(p, "%s", t);
            free(t);
            }
         *p++ = '@';
         }
      if( url->host != NULL )
         p += sprintf(p, "%s", url->host);
      if( url->port != 0 && url->port != HTTP_DEFAULT_PORT )
         p += sprintf(p, ":%d", url->port);
   case RELATIVE_URL:
      p += sprintf(p, "%s", url->path);
      if( qry != NULL ) {
         *p++ = '?';
         p += sprintf(p, "%s", qry);
         free(qry);
         }
      *p = '\0';
      break;
   default:
      free(u_str);
      return NULL;
   }

 return u_str;
} 


char *url_get_query_string(struct s_url *url)
{
 int i, ql;
 char *qs, *p, *t;
 int nsz[MAX_QUERYPAIR];
 int vsz[MAX_QUERYPAIR];

 if( url->query->sz == 0 )
    return NULL;

 for(ql=0, i=0; i<url->query->sz; i++) {
     nsz[i] = strlen(url->query->pair[i]->name) *3;
     vsz[i] = strlen(url->query->pair[i]->value) *3,
     ql += nsz[i] + vsz[i] +1;
     ql++;   /* add trailing '&' */
     }

 p = qs = xmalloc(ql);
 for(i=0; i<url->query->sz; i++) { 
     t = url_encode(url->query->pair[i]->name);
     p += sprintf(p, "%s", t);
     free(t);
     *p++ = '='; 
     t = url_encode(url->query->pair[i]->value);
     p += sprintf(p, "%s", t);
     free(t);
     *p++ = '&';
     }
 *(p-1) = '\0';

 return qs;
}
 

int url_recalibrate(struct s_url *new, struct s_url *orig)
{

 new->type = RECALIBRATED_URL;
 if( (new->scheme != 0) && (new->scheme != orig->scheme) ) {
    url_errno = UE_FORBIDDEN;
    return -1;
    }
 new->scheme = orig->scheme;
 if( new->user == NULL && orig->user != NULL )
    new->user = strdup(orig->user);
 if( new->pass == NULL && orig->pass != NULL )
    new->pass = strdup(orig->pass);
 if( new->host != NULL ) {
    if( strcasecmp(new->host, orig->host) != 0 ) {
       url_errno = UE_FORBIDDEN;
       return -1;
       }
    }
 else
    new->host = strdup(orig->host);
 if( new->port != 0 && orig->port != new->port ) {
    url_errno = UE_FORBIDDEN;
    return -1;
    }
 if( new->path[0] != '/' ) {      /* relative path */
    int szd, szs;
    char *p, *q, *t;
    p = orig->path;
    q = strrchr(p, '/'); /* impossible ? */
    if( q == NULL ) {
       url_errno = UE_UNDEFINED;
       return -1;      
       }
    szd = strlen(new->path);
    szs = (q-p)+1;
    t = xmalloc(szs+szd+1);
    strncpy(t, orig->path, szs);
    strncpy(t+szs, new->path, szd);
    t[szs+szd] = '\0';
    free(new->path);
    new->path = t;
    }

 return 0;
}


/*
 * for 30x code redirection,
 * rebuild from 'redir'
 */
struct s_url *url_redirect(struct s_url *url, char *redir)
{
 int szs, szd, i;
 char *s, *p, *t, *u, *sl;
 struct s_url *new_url;

 s = p = strdup(redir);
 if( strstr(redir, "://") == NULL ) {
    /*
     * welcome to a magic world where
     * the word RFC means nothing.
     */
    new_url = url_dup(url);
    free(new_url->path);
    for(i=0; i<url->query->sz; i++) {
        free(new_url->query->pair[i]->name);
        free(new_url->query->pair[i]->value);
        }
    if( *p != '/' ) {
       sl = strrchr(url->path, '/');
       if( sl == NULL ) {
          url_errno = UE_UNDEFINED;
          return NULL;
          }
       szd = (sl-(url->path))+1;
       szs = strlen(redir);
       u = t = xmalloc(szd+szs+1);
       strncpy(t, url->path, szd);
       strncpy(t+szd, redir, szs);
       t[szs+szd] = '\0';
       url_get_path(&t, new_url);
       url_get_query(&t, new_url);
       free(u);
       }
    else {
       url_get_path(&p, new_url);
       url_get_query(&p, new_url);
       }
    }
 else {
    new_url = url_build(redir);
    }

 free(s);
 return new_url;
}


struct s_url *url_dup(struct s_url *url)
{
 int i;
 struct s_url *n;

 n = url_alloc();
 n->method = url->method;
 n->type = url->type;
 n->scheme = url->scheme;
 if( url->user )
    n->user = strdup(url->user);
 if( url->pass )
    n->pass = strdup(url->pass);
 if( url->host )
    n->host = strdup(url->host);
 n->port = url->port;
 n->path = strdup(url->path);
 for(i=0; i<url->query->sz; i++) 
    url_querypair_add(url->query->pair[i]->name,
                      url->query->pair[i]->value,
                      i, n);
 if( url->frag )
    n->frag = strdup(url->frag);

 return n;
}


/*           *
 ** OPTIONS **
 *           */


int url_set_context(int n)
{
 
 url_context = n;

 return url_context;
}


int url_get_context()
{

 return url_context;
}


/*                  *
 ** URL LIST BUILD **
 *                  */


void url_list_urladd(struct url_list *ul, struct s_url *url)
{
 
 if( ul->head == NULL ) {
    ul->head = e_url_alloc();
    ul->head->url  = url;
    ul->tail = ul->head;
    }
 else {
    ul->tail->next = e_url_alloc();
    ul->tail->next->url  = url;
    ul->tail = ul->tail->next;
    }
 ul->sz++;
}


int url_list_stradd(struct url_list *ul, char *strurl)
{
 struct s_url *n_url;

 n_url = url_build(strurl);
 if( n_url == NULL ) {
    m_warn("%s", url_strerror(url_errno));
    return -1;
    }

 url_list_urladd(ul, n_url);

 return 0;
}
